#-------------------------------------------------------------------------------------------------------------
#                 R code to generate the simulated data set for
#
#   Haplotype-based genome-wide prediction models exploit local epistatic interactions among markers
#
#                Yong Jiang, Renate H. Schmidt and Jochen C. Reif
#-------------------------------------------------------------------------------------------------------------




####-------------------------------------------------------------------------------------------------------------
#
#        Loading necessary functions
#
####-------------------------------------------------------------------------------------------------------------



## The function sampling QTL

QTLsampling <- function(M,X,p){
                   result <- list()
                   X$chr <- as.factor(X$chr)
                   NumChr <- nlevels(X$chr)
                   NumMarEachChr <- table(X$chr)
                   CumNumMar <- c(0,cumsum(NumMarEachChr))
                   submarind <- NULL
                   for (i in 1:NumChr)
                     {
                       submarind <- c(submarind,sort(sample((CumNumMar[i]+1):CumNumMar[i+1],p,replace=FALSE)))   
                     }
                   result$subMKdata <- M[,submarind]
                   result$subMPdata <- X[submarind,]
                   result$nmar <- p*NumChr
                   return(result)
                }

## The function generating coefficient matrix including higher-order epistasis

EPmatGen <- function(X){n <- nrow(X)
                        p <- ncol(X)
                        M <- X
                        if (p >= 2) 
                         {
                           for (j in 1:(p-1))
                             {
                               M <- cbind(M,matrix(rep(X[,j],p-j),nrow=n)*X[,(j+1):p])
                             } 
                         } 
                        if (p == 3)
                         {
                           M <- cbind(M,X[,1]*X[,2]*X[,3])
                         }
                        if (p == 4)
                         {
                           M <- cbind(M,X[,1]*X[,2]*X[,3])
                           M <- cbind(M,X[,1]*X[,2]*X[,4])
                           M <- cbind(M,X[,1]*X[,3]*X[,4])
                           M <- cbind(M,X[,2]*X[,3]*X[,4])
                           M <- cbind(M,X[,1]*X[,2]*X[,3]*X[,4])
                         }
                        if (p == 5)
                         {
                           M <- cbind(M,X[,1]*X[,2]*X[,3])
                           M <- cbind(M,X[,1]*X[,2]*X[,4])
                           M <- cbind(M,X[,1]*X[,2]*X[,5])
                           M <- cbind(M,X[,1]*X[,3]*X[,4])
                           M <- cbind(M,X[,1]*X[,3]*X[,5])
                           M <- cbind(M,X[,1]*X[,4]*X[,5])
                           M <- cbind(M,X[,2]*X[,3]*X[,4])
                           M <- cbind(M,X[,2]*X[,3]*X[,5])
                           M <- cbind(M,X[,2]*X[,4]*X[,5])
                           M <- cbind(M,X[,3]*X[,4]*X[,5])
                           M <- cbind(M,X[,1]*X[,2]*X[,3]*X[,4])
                           M <- cbind(M,X[,1]*X[,2]*X[,3]*X[,5])
                           M <- cbind(M,X[,1]*X[,2]*X[,4]*X[,5])
                           M <- cbind(M,X[,1]*X[,3]*X[,4]*X[,5])
                           M <- cbind(M,X[,2]*X[,3]*X[,4]*X[,5])
                           M <- cbind(M,X[,1]*X[,2]*X[,3]*X[,4]*X[,5])
                         }
                        return(M)
                       }



## The function generating covariance matrices

COVGEN <- function(r){
              result <- list()
              result$cormat <- list()
              result$fullcormat <- list()
              result$desmat <- list()
              result$invdesmat <- list()
              result$V <- list()
              result$fullV <- list()
              result$lambda <- list()
              result$fulllambda <- list()

              result$cormat[[1]] <- 1
              result$V[[1]] <- 1
              result$desmat[[1]] <- 1
              result$invdesmat[[1]] <- 1
              result$lambda[[1]] <- 1
              result$fullcormat[[1]] <- 1

              A <- matrix(0,3,3)
              A[,1] <- c(2,2,0)
              A[,2] <- c(2,0,2)
              A[,3] <- A[,1]*A[,2]
              result$desmat[[2]] <- A 
              result$invdesmat[[2]] <- solve(A)
              result$cormat[[2]] <- 4*solve(A)%*%t(solve(A))
              result$fullcormat[[2]] <- result$cormat[[2]]

              A <- matrix(0,7,7)
              A[,1] <- c(2,2,2,2,0,0,0)
              A[,2] <- c(2,2,0,0,2,2,0)
              A[,3] <- c(2,0,2,0,2,0,2)
              A[,4] <- A[,1]*A[,2]
              A[,5] <- A[,1]*A[,3]
              A[,6] <- A[,2]*A[,3]
              A[,7] <- A[,1]*A[,2]*A[,3]
              result$desmat[[3]] <- A
              result$invdesmat[[3]] <- solve(A)
              B <- 4*solve(A)%*%t(solve(A))
              result$cormat[[3]] <- B[1:6,1:6]
              result$fullcormat[[3]] <- B

              A <- matrix(0,15,15)
              A[,1] <- c(2,2,2,2,2,2,2,2,0,0,0,0,0,0,0)
              A[,2] <- c(2,2,2,2,0,0,0,0,2,2,2,2,0,0,0)
              A[,3] <- c(2,2,0,0,2,2,0,0,2,2,0,0,2,2,0)
              A[,4] <- c(2,0,2,0,2,0,2,0,2,0,2,0,2,0,2)
              A[,5] <- A[,1]*A[,2]
              A[,6] <- A[,1]*A[,3]
              A[,7] <- A[,1]*A[,4]
              A[,8] <- A[,2]*A[,3]
              A[,9] <- A[,2]*A[,4]
              A[,10] <- A[,3]*A[,4]
              A[,11] <- A[,1]*A[,2]*A[,3]
              A[,12] <- A[,1]*A[,2]*A[,4]
              A[,13] <- A[,1]*A[,3]*A[,4]
              A[,14] <- A[,2]*A[,3]*A[,4]
              A[,15] <- A[,1]*A[,2]*A[,3]*A[,4]
              result$desmat[[4]] <- A
              result$invdesmat[[4]] <- solve(A)
              B <- 4*solve(A)%*%t(solve(A))
              result$cormat[[4]] <- B[1:10,1:10]
              result$fullcormat[[4]] <- B

              A <- matrix(0,31,31)
              A[,1] <- c(2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0)
              A[,2] <- c(2,2,2,2,2,2,2,2,0,0,0,0,0,0,0,0,2,2,2,2,2,2,2,2,0,0,0,0,0,0,0)
              A[,3] <- c(2,2,2,2,0,0,0,0,2,2,2,2,0,0,0,0,2,2,2,2,0,0,0,0,2,2,2,2,0,0,0)
              A[,4] <- c(2,2,0,0,2,2,0,0,2,2,0,0,2,2,0,0,2,2,0,0,2,2,0,0,2,2,0,0,2,2,0)
              A[,5] <- c(2,0,2,0,2,0,2,0,2,0,2,0,2,0,2,0,2,0,2,0,2,0,2,0,2,0,2,0,2,0,2)
              A[,6] <- A[,1]*A[,2]
              A[,7] <- A[,1]*A[,3]
              A[,8] <- A[,1]*A[,4]
              A[,9] <- A[,1]*A[,5]
              A[,10] <- A[,2]*A[,3]
              A[,11] <- A[,2]*A[,4]
              A[,12] <- A[,2]*A[,5]
              A[,13] <- A[,3]*A[,4]
              A[,14] <- A[,3]*A[,5]
              A[,15] <- A[,4]*A[,5]
              A[,16] <- A[,1]*A[,2]*A[,3]
              A[,17] <- A[,1]*A[,2]*A[,4]
              A[,18] <- A[,1]*A[,2]*A[,5]
              A[,19] <- A[,1]*A[,3]*A[,4]
              A[,20] <- A[,1]*A[,3]*A[,5]
              A[,21] <- A[,1]*A[,4]*A[,5]
              A[,22] <- A[,2]*A[,3]*A[,4]
              A[,23] <- A[,2]*A[,3]*A[,5]
              A[,24] <- A[,2]*A[,4]*A[,5]
              A[,25] <- A[,3]*A[,4]*A[,5]
              A[,26] <- A[,1]*A[,2]*A[,3]*A[,4]
              A[,27] <- A[,1]*A[,2]*A[,3]*A[,5]
              A[,28] <- A[,1]*A[,2]*A[,4]*A[,5]
              A[,29] <- A[,1]*A[,3]*A[,4]*A[,5]
              A[,30] <- A[,2]*A[,3]*A[,4]*A[,5]
              A[,31] <- A[,1]*A[,2]*A[,3]*A[,4]*A[,5]
              result$desmat[[5]] <- A
              result$invdesmat[[5]] <- solve(A)
              B <- 4*solve(A)%*%t(solve(A))
              result$cormat[[5]] <- B[1:10,1:10]
              result$fullcormat[[5]] <- B

              for (i in 2:5)
                {
                  result$V[[i]] <- eigen(result$cormat[[i]])$vectors
                  result$lambda[[i]] <- eigen(result$cormat[[i]])$values
                }

              for (i in 2:5)
                {
                  result$fullV[[i]] <- eigen(result$fullcormat[[i]])$vectors
                  result$fulllambda[[i]] <- eigen(result$fullcormat[[i]])$values
                }

              if (r==3/4) 
                { 
                  return(result) 
                }else{
                  newresult <- list()
                  newresult$desmat <- result$desmat
                  newresult$invdesmat <- result$invdesmat
                  newresult$fullcormat <- list()
                  newresult$fullV <- list()
                  newresult$fulllambda <- list()
                  
                  newresult$fullcormat[[1]] <- 1
                  for (i in 2:5)
                    { 
                      newresult$fullcormat[[i]] <- matrix(0,2^i-1,2^i-1)
                      newresult$fullcormat[[i]][1:i,1:i] <- result$fullcormat[[i]][1:i,1:i]
                      newresult$fullcormat[[i]][(i+1):(2^i-1),1:i] <- sqrt((r/(3/4)))*result$fullcormat[[i]][(i+1):(2^i-1),1:i]
                      newresult$fullcormat[[i]][1:i,(i+1):(2^i-1)] <- sqrt((r/(3/4)))*result$fullcormat[[i]][1:i,(i+1):(2^i-1)]
                      newresult$fullcormat[[i]][(i+1):(2^i-1),(i+1):(2^i-1)] <- (r/(3/4))*result$fullcormat[[i]][(i+1):(2^i-1),(i+1):(2^i-1)]
                    }

                  newresult$fullV[[1]] <- 1
                  newresult$fulllambda[[1]] <- 1
                  for (i in 2:5)
                    {
                      newresult$fullV[[i]] <- eigen(newresult$fullcormat[[i]])$vectors
                      newresult$fulllambda[[i]] <- eigen(newresult$fullcormat[[i]])$values
                    }
                  return (newresult)
                } # end if

             } # end function



## The functions taking markers into haplotype alleles

MarToHap <- function(x){w <- length(x)
                        h <- NULL
                        for (k in 1:w)
                          {
                            if (x[k]==2)
                              {
                                h <- paste0(h,"1")
                              }else{
                                if(x[k]==1)
                                  {
                                    h <- c(paste0(h,"0"),paste0(h,"1"))
                                  }else{
                                    h <- paste0(h,"0")
                                  }
                              }
                          }
                        return(h)
                       }

HapCoeff <- function(x,hapalleles){z <- array(0,length(hapalleles))
                        names(z) <- hapalleles
                        z[x] <- 2/length(x)
                        return(z)
                       }

hapal <- list()
hapal[[1]] <- c("1","0")
hapal[[2]] <- paste0(c(rep("1",2),rep("0",2)),rep(hapal[[1]],2))
hapal[[3]] <- paste0(c(rep("1",4),rep("0",4)),rep(hapal[[2]],2))
hapal[[4]] <- paste0(c(rep("1",8),rep("0",8)),rep(hapal[[3]],2))
hapal[[5]] <- paste0(c(rep("1",16),rep("0",16)),rep(hapal[[4]],2))



## The main function of simulation in scenario 1 (no epistasis)

SIMTRAIT1 <- function(M,X,p,h2,n){
                       result <- list()
                       ngen <- nrow(M)
                       nsim <- n 
                       QTL <- QTLsampling(M,X,p)
                       nmar <- QTL$nmar
                       simtrait <- matrix(0,ngen,nsim)
                       rownames(simtrait) <- rownames(QTL$subMKdata)
                       colnames(simtrait) <- paste0("SimTrait_",1:nsim)
                       simTrueGV <- simtrait

                       for (k in 1:nsim)
                         {
                           x <- rnorm(nmar,0,1)
                           simTrueGV[,k] <- QTL$subMKdata%*%x
                           varg <- var(simTrueGV[,k])
                           vare <- (1-h2)/h2*varg
                           simtrait[,k] <- simTrueGV[,k]+rnorm(ngen,0,sqrt(vare))
                         }# end k
    
                       result$simtrait <- simtrait
                       result$simTrueGV <- simTrueGV
                       return(result)
                     }


## The main function of simulation in scenario 2 (global epistasis)

SIMTRAIT2 <- function(M,X,p,h2,r,n){
                       result <- list()
                       ngen <- nrow(M)
                       nsim <- n 
                       QTL <- QTLsampling(M,X,p)
                       nmar <- QTL$nmar
                       subMKdata <- QTL$subMKdata

                       simtrait <- matrix(0,ngen,nsim)
                       rownames(simtrait) <- rownames(subMKdata)
                       colnames(simtrait) <- paste0("SimTrait_",1:nsim)
                       simTrueGV <- simtrait

                       for (k in 1:nsim)
                         {
                           EPmat <- NULL
                           for (j in 1:(p-1))
                             {
                               EPmat <- cbind(EPmat,matrix(rep(subMKdata[,j],p-j),nrow=ngen)*subMKdata[,(j+1):p])
                             }
                           x <- rnorm(nmar,0,1)
                           y <- rnorm(ncol(EPmat),0,sqrt(r))
                           simTrueGV[,k] <- subMKdata%*%x+EPmat%*%y
                           varg <- var(simTrueGV[,k])
                           vare <- (1-h2)/h2*varg
                           simtrait[,k] <- simTrueGV[,k]+rnorm(ngen,0,sqrt(vare))
                         }

                       result$simtrait <- simtrait
                       result$simTrueGV <- simTrueGV
                       return(result)
                     }


## The main function of simulation in scenario 3 (independent di-genic local epistasis)


SIMTRAIT3 <- function(M,X,p,h2,r,n){
                       result <- list()
                       ngen <- nrow(M)
                       nsim <- n 
                       QTL <- QTLsampling(M,X,p)
                       nmar <- QTL$nmar
                       subMKdata <- QTL$subMKdata

                       simtrait <- matrix(0,ngen,nsim)
                       rownames(simtrait) <- rownames(subMKdata)
                       colnames(simtrait) <- paste0("SimTrait_",1:nsim)
                       simTrueGV <- simtrait

                       for (m in 1:nsim)
                         { 
                           a <- 0
                           EPmat <- NULL
                           for (k in 1:p)
                             {
                               simw <- sample(2:5,1)
                               pos <- (a+1):min(a+simw,p)
                               localM <- subMKdata[,pos,drop=FALSE]
                               if (ncol(localM) > 1)
                                 {
                                   for (j in 1:(ncol(localM)-1))
                                     {
                                       EPmat <- cbind(EPmat,matrix(rep(localM[,j],ncol(localM)-j),nrow=ngen)*localM[,(j+1):ncol(localM)])
                                     }
                                 }
                               a <- min(a+simw,100)
                               if (a==p) {break}
                             }  
                           x <- rnorm(nmar,0,1)
                           y <- rnorm(ncol(EPmat),0,sqrt(r))
                           simTrueGV[,m] <- subMKdata%*%x+EPmat%*%y
                           varg <- var(simTrueGV[,m])
                           vare <- (1-h2)/h2*varg
                           simtrait[,m] <- simTrueGV[,m]+rnorm(ngen,0,sqrt(vare))
                         }

                       result$simtrait <- simtrait
                       result$simTrueGV <- simTrueGV
                       return(result)
                     }



## The main function of simulation in scenario 4 (independent full local epistasis)

SIMTRAIT4 <- function(M,X,p,h2,r,n){
                       result <- list()
                       ngen <- nrow(M)
                       nsim <- n 
                       QTL <- QTLsampling(M,X,p)
                       nmar <- QTL$nmar
                       subMKdata <- QTL$subMKdata

                       simtrait <- matrix(0,ngen,nsim)
                       rownames(simtrait) <- rownames(subMKdata)
                       colnames(simtrait) <- paste0("SimTrait_",1:nsim)
                       simTrueGV <- simtrait

                       covres <- COVGEN(r)
                       fullcormat <- covres$fullcormat
                       fulllambda <- covres$fulllambda
                       fullV <- covres$fullV
                       for (m in 1:nsim)
                         {    
                           a <- 0
                           for (k in 1:p)
                             {
                               simw <- sample(2:5,1)
                               pos <- (a+1):min(a+simw,p)
                               localM <- subMKdata[,pos,drop=FALSE]
                               if (ncol(localM) > 1)
                                 {
                                   d <- ncol(localM)
                                   x <- rnorm(nrow(fullcormat[[d]]),0,sqrt(fulllambda[[d]]))
                                   y <- as.vector(fullV[[d]]%*%x)
                                   EPmat <- as.matrix(EPmatGen(localM))
                                   simTrueGV[,m] <- simTrueGV[,m] + EPmat%*%y
                                 }
                               a <- min(a+simw,p)
                               if (a==p) {break}
                             }  
                           varg <- var(simTrueGV[,m])
                           vare <- (1-h2)/h2*varg
                           simtrait[,m] <- simTrueGV[,m]+rnorm(ngen,0,sqrt(vare))
                         }

                       result$simtrait <- simtrait
                       result$simTrueGV <- simTrueGV
                       return(result)
                     }




## The main function of simulation in scenario 5 (correlated di-genic local epistasis)

M <- MKdata
X <- MPdata
p <- 100
h2 <- 0.7
r <- 3/4
n <- 10

SIMTRAIT5 <- function(M,X,p,h2,r,n){
                result <- list()
                ngen <- nrow(M)
                nsim <- n 
                QTL <- QTLsampling(M,X,p)
                nmar <- QTL$nmar
                subMKdata <- QTL$subMKdata
                subMPdata <- QTL$subMPdata

                subMPdata$chr <- as.factor(subMPdata$chr)
                nchr <- nlevels(subMPdata$chr)
                NumMarEachChr <- table(subMPdata$chr) 
                CumNumMar <- c(0,cumsum(NumMarEachChr))

                simtrait <- matrix(0,ngen,nsim)
                rownames(simtrait) <- rownames(subMKdata)
                colnames(simtrait) <- paste0("SimTrait_",1:nsim)
                simTrueGV <- simtrait

                xx <- c(1,3,6,10,15)
                va <- c(1,3/4,7/16,15/64,31/256)

                covres <- COVGEN(r)
                if (r==3/4)
                  {
                    # standard ratio of variances
                    for (m in 1:nsim)
                      {     
                        for (k in 1:nchr)
                          {                              
                            subX <- subMKdata[,(CumNumMar[k]+1):CumNumMar[k+1],drop=FALSE]
                            a <- 0
                            for (l in 1:p)
                              {
                                simw <- sample(2:5,1)
                                pos <- (a+1):min(a+simw,p)
                                subsubX <- as.matrix(subX[,pos,drop=FALSE])
                                w <- length(pos)
                                if (ncol(subsubX)==1)
                                  {
                                    subHapMat <- subsubX
                                    x <- rnorm(ncol(subHapMat),0,1)
                                    simTrueGV[,m] <- simTrueGV[,m]+subHapMat%*%x
                                  }else{
                                    AlleleList <- apply(subsubX,1,MarToHap)
                                    hapalleles <- unique(unlist(AlleleList))
                                    subHapMat <- t(sapply(AlleleList,HapCoeff,hapalleles))
                                    rownames(subHapMat) <- rownames(subsubX)
                                    tpos <- which(colnames(subHapMat)==hapal[[w]][2^w])
                                    if (length(tpos)>0)
                                      {
                                        subHapMat <- subHapMat[,-tpos,drop=FALSE]  
                                      }              

                                    poss <- NULL
                                    for (j in 1:ncol(subHapMat))
                                      {
                                        poss <- c(poss,which(hapal[[w]]==colnames(subHapMat)[j]))
                                      }
                                    W <- covres$invdesmat[[w]][,poss]
                                    V <- 2*W%*%diag(1,2^w-1)[poss,]
                                    vcov <- V%*%t(V)
                
                                    lambda <- eigen(vcov)$values
                                    U <- eigen(vcov)$vectors
                                    len <- length(which(lambda<0.00001))
                                    lambda[which(lambda<0.00001)] <- 0
                                    x <- rnorm(length(lambda),0,sqrt(lambda))
                                    y <- as.vector(U%*%x)
                                    EPmat <- NULL
                                    for (j in 1:(ncol(subsubX)-1))
                                      {
                                        EPmat <- cbind(EPmat,matrix(rep(subsubX[,j],ncol(subsubX)-j),nrow=ngen)*subsubX[,(j+1):ncol(subsubX)])
                                      }
                                    simTrueGV[,m] <- simTrueGV[,m] + cbind(subsubX,EPmat)%*%y[1:(ncol(EPmat)+w)]
                                  } # end if ncol(subsubX)                                   
                                a <- min(a+simw,p)
                                if (a==p) {break}
                              }# end of l
                          }# end of k     
                        varg <- var(simTrueGV[,m])
                        vare <- (1-h2)/h2*varg
                        simtrait[,m] <- simTrueGV[,m]+rnorm(ngen,0,sqrt(vare))
                      }# end of m

                  }else{
                    
                    # non-stardard ratio of variances  
                    for (m in 1:nsim)
                      {     
                        for (k in 1:nchr)
                          {                          
                            subX <- subMKdata[,(CumNumMar[k]+1):CumNumMar[k+1],drop=FALSE]
                            a <- 0
                            for (l in 1:p)
                              {
                                simw <- sample(2:5,1)
                                pos <- (a+1):min(a+simw,p)
                                subsubX <- as.matrix(subX[,pos,drop=FALSE])
                                w <- length(pos)
                                if (ncol(subsubX)==1)
                                  {
                                    subHapMat <- subsubX
                                    x <- rnorm(ncol(subHapMat),0,1)
                                    simTrueGV[,m] <- simTrueGV[,m]+subHapMat%*%x
                                  }else{
                                    AlleleList <- apply(subsubX,1,MarToHap)
                                    hapalleles <- unique(unlist(AlleleList))
                                    subHapMat <- t(sapply(AlleleList,HapCoeff,hapalleles))
                                    rownames(subHapMat) <- rownames(subsubX)
                                    tpos <- which(colnames(subHapMat)==hapal[[w]][2^w])
                                    if (length(tpos)>0)
                                      {
                                        subHapMat <- subHapMat[,-tpos,drop=FALSE]  
                                      }              
                                    poss <- NULL
                                    for (j in 1:ncol(subHapMat))
                                      {
                                        poss <- c(poss,which(hapal[[w]]==colnames(subHapMat)[j]))
                                      }
                                    W <- covres$invdesmat[[w]][,poss]
                                    V <- 2*W%*%diag(1,2^w-1)[poss,]
                                    vcov <- V%*%t(V)
                                    vcov <- vcov[1:xx[w],1:xx[w]]
                                    vcov <- vcov*sqrt(r/(3/4))
                                    diag(vcov) <- diag(vcov)/sqrt(r/(3/4))
                                    diag(vcov)[-1:-w] <- diag(vcov)[-1:-w]*(r/(3/4))
                                
                                    lambda <- eigen(vcov)$values
                                    U <- eigen(vcov)$vectors
                                    len <- length(which(lambda<0.00001))
                                    lambda[which(lambda<0.00001)] <- 0
                                    x <- rnorm(length(lambda),0,sqrt(lambda))
                                    y <- as.vector(U%*%x)
                                    EPmat <- NULL
                                    for (j in 1:(ncol(subsubX)-1))
                                      {
                                        EPmat <- cbind(EPmat,matrix(rep(subsubX[,j],ncol(subsubX)-j),nrow=ngen)*subsubX[,(j+1):ncol(subsubX)])
                                      }
                                    simTrueGV[,m] <- simTrueGV[,m] + cbind(subsubX,EPmat)%*%y
                                  }                                    
                                a <- min(a+simw,p)
                                if (a==p) {break} 
                              }# end of l
                          }# end of k    
                        varg <- var(simTrueGV[,m])
                        vare <- (1-h2)/h2*varg
                        simtrait[,m] <- simTrueGV[,m]+rnorm(ngen,0,sqrt(vare))
                      }# end of m

                   } # end if r==3/4
                
                 result$simtrait <- simtrait
                 result$simTrueGV <- simTrueGV
                 return(result)
               }# end function




## The main function of simulation in scenario 6 (correlated full local epistasis)


SIMTRAIT6 <- function(M,X,p,h2,r,n){
                result <- list()
                ngen <- nrow(M)
                nsim <- n 
                QTL <- QTLsampling(M,X,p)
                nmar <- QTL$nmar
                subMKdata <- QTL$subMKdata
                subMPdata <- QTL$subMPdata
             
                subMPdata$chr <- as.factor(subMPdata$chr)
                nchr <- nlevels(subMPdata$chr)
                NumMarEachChr <- table(subMPdata$chr) 
                CumNumMar <- c(0,cumsum(NumMarEachChr))

                simtrait <- matrix(0,ngen,nsim)
                rownames(simtrait) <- rownames(subMKdata)
                colnames(simtrait) <- paste0("SimTrait_",1:nsim)
                simTrueGV <- simtrait

                if (r==3/4)
                  {
                    
                    # standard ratio of variances as derived in HGBLUP  
                    for (m in 1:nsim)
                      {    
                        for (k in 1:nchr)
                          {                                                              
                            subX <- subMKdata[,(CumNumMar[k]+1):CumNumMar[k+1],drop=FALSE]
                            a <- 0
                            for (l in 1:p)
                              {
                                simw <- sample(2:5,1)
                                pos <- (a+1):min(a+simw,p)
                                subsubX <- as.matrix(subX[,pos,drop=FALSE])
                                if (ncol(subsubX)==1)
                                  {
                                    subHapMat <- subsubX
                                  }else{
                                    AlleleList <- apply(subsubX,1,MarToHap)
                                    hapalleles <- unique(unlist(AlleleList))
                                    subHapMat <- t(sapply(AlleleList,HapCoeff,hapalleles))
                                    rownames(subHapMat) <- rownames(subsubX)
                                  }                                    
                                x <- rnorm(ncol(subHapMat),0,1)
                                simTrueGV[,m] <- simTrueGV[,m]+subHapMat%*%x
                                a <- min(a+simw,p)
                                if (a==p) {break}
                              }# end of l
                          }# end of k
        
                        varg <- var(simTrueGV[,m])
                        vare <- (1-h2)/h2*varg
                        simtrait[,m] <- simTrueGV[,m]+rnorm(ngen,0,sqrt(vare))
                      }# end of m
 
                  }else{

                    # non-standard ratio of variances
                    covres <- COVGEN(r) 
                    for (m in 1:nsim)
                      {      
                        for (k in 1:nchr)
                          {                               
                            subX <- subMKdata[,(CumNumMar[k]+1):CumNumMar[k+1],drop=FALSE]
                            a <- 0
                            for (l in 1:p)
                              {
                               simw <- sample(2:5,1)
                               pos <- (a+1):min(a+simw,p)
                               subsubX <- as.matrix(subX[,pos,drop=FALSE])
                               w <- length(pos)
                               if (ncol(subsubX)==1)
                                 {
                                   subHapMat <- subsubX
                                   x <- rnorm(ncol(subHapMat),0,1)
                                   simTrueGV[,m] <- simTrueGV[,m]+subHapMat%*%x
                                 }else{
                                   AlleleList <- apply(subsubX,1,MarToHap)
                                   hapalleles <- unique(unlist(AlleleList))
                                   subHapMat <- t(sapply(AlleleList,HapCoeff,hapalleles))
                                   rownames(subHapMat) <- rownames(subsubX)
                                   tpos <- which(colnames(subHapMat)==hapal[[w]][2^w])
                                   if (length(tpos)>0)
                                     {
                                       subHapMat <- subHapMat[,-tpos,drop=FALSE]  
                                     }              
                                   poss <- NULL
                                   for (j in 1:ncol(subHapMat))
                                     {
                                       poss <- c(poss,which(hapal[[w]]==colnames(subHapMat)[j]))
                                     }
                                   W <- covres$invdesmat[[w]][,poss]
                                   V <- 2*W%*%diag(1,2^w-1)[poss,]
                                   vcov <- V%*%t(V)
                                   newvcov <- vcov
                                   newvcov[1:w,1:w] <- vcov[1:w,1:w]
                                   newvcov[(w+1):(2^w-1),1:w] <- sqrt(r/(3/4))*vcov[(w+1):(2^w-1),1:w]
                                   newvcov[1:w,(w+1):(2^w-1)] <- sqrt(r/(3/4))*vcov[1:w,(w+1):(2^w-1)] 
                                   newvcov[(w+1):(2^w-1),(w+1):(2^w-1)] <- (r/(3/4))*vcov[(w+1):(2^w-1),(w+1):(2^w-1)]
                                
                                   lambda <- eigen(newvcov)$values
                                   U <- eigen(newvcov)$vectors
                                   len <- length(which(lambda<0.00001))
                                   lambda[which(lambda<0.00001)] <- 0
                                   x <- rnorm(length(lambda),0,sqrt(lambda))
                                   y <- as.vector(U%*%x)
                                   EPmat <- as.matrix(EPmatGen(subsubX))
                                   simTrueGV[,m] <- simTrueGV[,m] + EPmat%*%y
                                 } # end if ncol(subsubX)                                    
                               a <- min(a+simw,p)
                               if (a==p) {break}
                             }# end of l
                          }# end of k     
                       varg <- var(simTrueGV[,m])
                       vare <- (1-h2)/h2*varg
                       simtrait[,m] <- simTrueGV[,m]+rnorm(ngen,0,sqrt(vare))
                     }# end of m

                  } # end if r==3/4

                result$simtrait <- simtrait
                result$simTrueGV <- simTrueGV
                return(result)
               } # end function


## The main function of simulation


SIMTRAIT <- function(M,X,p,h2,r,s,n){
                       if (s==1){ result <- SIMTRAIT1(M,X,p,h2,n) }
                       if (s==2){ result <- SIMTRAIT2(M,X,p,h2,r,n) }
                       if (s==3){ result <- SIMTRAIT3(M,X,p,h2,r,n) }
                       if (s==4){ result <- SIMTRAIT4(M,X,p,h2,r,n) }
                       if (s==5){ result <- SIMTRAIT5(M,X,p,h2,r,n) }
                       if (s==6){ result <- SIMTRAIT6(M,X,p,h2,r,n) }
                       if (!(s %in% 1:6)) { cat("Error: Incorrect s value!\n") }
                       return(result)
                     }



# End of loading functions
####-------------------------------------------------------------------------------------------------------------
####-------------------------------------------------------------------------------------------------------------






###-------------------------------------------------------------------------------------------------------------
#
#                The body code
#
###-------------------------------------------------------------------------------------------------------------


# load the genomic data
# after downloading the data, the genomic data must be organized as follows
# rows correspond to individuals
# columns corresond to SNP markers
# SNP markers must be coded as 0, 1, 2
MKdata <- read.table("Sample_genomic_data.txt",header=TRUE)
MKdata <- as.matrix(MKdata)

# load the genetic/physical map data
# after downloading the data, the file must be organized as follows
# rows correspond to SNP markers
# the first column consists of the chromosome information for each marker
# the second column contains the position information for each marker
MPdata <- read.table("Sample_map_data.txt",header=TRUE)
MPdata$chr <- as.factor(MPdata$chr)


# start simulation using the function SIMTRAIT
# scenario 1: only additive
# scenario 2: global epistasis
# scenario 3: independent di-genic local epistasis
# scenario 4: independent full local epistasis 
# scenario 5: correlated di-genic local epistasis
# scenario 6: correlated full local epistasis 

simresult <- SIMTRAIT(M=MKdata,   # a matrix of genomic data
                X=MPdata,   # a dataframe of genetic/physical map (as MPdata)
                p=100,      # the number of markers sampled from each chromosome as simulated QTL, default 100
                h2=0.7,     # the simulated heritability, default 0.7
                r=3/4,      # the simulated ratio of var_a/var_aa, default 3/4
                s=1,        # the simulated scenario, taking integral values 1 to 6
                n=10        # the number of simulated traits (replications)
                )

simresult$simtrait  # this is the simulated traits

simresult$simTrueGV  # this is the simulated true genetic values














































